package ru.dbotthepony.mc.otm.datagen

import com.mojang.serialization.JsonOps
import net.minecraft.core.Holder
import net.minecraft.core.HolderOwner
import net.minecraft.core.registries.Registries
import net.minecraft.resources.RegistryOps
import net.minecraft.resources.ResourceKey
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.packs.PackType
import net.minecraft.tags.BlockTags
import net.minecraft.world.level.levelgen.VerticalAnchor
import net.minecraft.world.level.levelgen.feature.ConfiguredFeature
import net.minecraft.world.level.levelgen.feature.Feature
import net.minecraft.world.level.levelgen.feature.configurations.OreConfiguration
import net.minecraft.world.level.levelgen.placement.CountPlacement
import net.minecraft.world.level.levelgen.placement.HeightRangePlacement
import net.minecraft.world.level.levelgen.placement.InSquarePlacement
import net.minecraft.world.level.levelgen.placement.PlacedFeature
import net.minecraft.world.level.levelgen.structure.templatesystem.TagMatchTest
import net.minecraftforge.common.data.JsonCodecProvider
import net.minecraftforge.data.event.GatherDataEvent
import ru.dbotthepony.mc.otm.registry.MBlocks

private object Owner : HolderOwner<Nothing> {
	override fun canSerializeIn(p_255875_: HolderOwner<Nothing>): Boolean {
		return true // because we can always reference things, even if they are not "registered" at this moment
	}
}

private data class Both<T>(val key: Holder<T>, val value: Holder<T>)

private fun <T> createHolders(key: ResourceKey<T>, value: T): Both<T> {
	val keyOnly = Holder.Reference.createStandAlone(Owner as HolderOwner<T>, key)
	val valueOnly = Holder.direct(value)
	return Both(keyOnly, valueOnly)
}

private fun <T> remap(vararg input: Both<T>): Map<ResourceLocation, Holder<T>> {
	val result = LinkedHashMap<ResourceLocation, Holder<T>>()

	for (value in input) {
		result[value.key.unwrapKey().get().location()] = value.value
	}

	return result
}

fun registerOreGen(event: GatherDataEvent) {
	val stone = TagMatchTest(BlockTags.STONE_ORE_REPLACEABLES)
	val deepslate = TagMatchTest(BlockTags.DEEPSLATE_ORE_REPLACEABLES)

	val target by lazy {
		listOf(
			OreConfiguration.target(stone, MBlocks.TRITANIUM_ORE.defaultBlockState()),
			OreConfiguration.target(deepslate, MBlocks.DEEPSLATE_TRITANIUM_ORE.defaultBlockState()),
		)
	}

	val ops = RegistryOps.create(JsonOps.INSTANCE, event.lookupProvider.join())
	val tritanium = createHolders(ResourceKey.create(Registries.CONFIGURED_FEATURE, modLocation("ore_tritanium")), ConfiguredFeature(Feature.ORE, OreConfiguration(target, 9)))
	val configuredFeatures = remap(tritanium)

	event.generator.addProvider(event.includeServer(), JsonCodecProvider(
		event.generator.packOutput, event.existingFileHelper, DataGen.MOD_ID, ops,
		PackType.SERVER_DATA, "worldgen/configured_feature", ConfiguredFeature.CODEC, configuredFeatures
	))

	val placedTritanium = createHolders(ResourceKey.create(Registries.PLACED_FEATURE, modLocation("normal_tritanium")), PlacedFeature(
		tritanium.key,
		listOf(
			CountPlacement.of(8),
			InSquarePlacement.spread(),
			HeightRangePlacement.triangle(VerticalAnchor.absolute(0), VerticalAnchor.absolute(50))
		)
	))

	val placedTritaniumDeep = createHolders(ResourceKey.create(Registries.PLACED_FEATURE, modLocation("deep_tritanium")), PlacedFeature(
		tritanium.key,
		listOf(
			CountPlacement.of(10),
			InSquarePlacement.spread(),
			HeightRangePlacement.uniform(VerticalAnchor.aboveBottom(8), VerticalAnchor.absolute(0))
		)
	))

	val placedFeatures = remap(placedTritanium, placedTritaniumDeep)

	event.generator.addProvider(event.includeServer(), JsonCodecProvider(
		event.generator.packOutput, event.existingFileHelper, DataGen.MOD_ID, ops,
		PackType.SERVER_DATA, "worldgen/placed_feature", PlacedFeature.CODEC, placedFeatures
	))
}
